/*
 * nam - turn "nroff -man|col|expand|fontch" output back
 *	into "nroff -man" input
 *
 * knowledge of the man macros and nroff (input and output) is scattered
 * throughout, but in general this approach should be applicable for any
 * formatter with simple input and rigid output formats.
 */
/*
 * Copyright (c) Geoffrey Collyer 1990, 1991, 1992, 1993, 1994.
 * All rights reserved.
 * Written by Geoffrey Collyer.
 * This software is not subject to any license of the American Telephone
 * and Telegraph Company or of the Regents of the University of California.
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. The authors are not responsible for the consequences of use of this
 *    software, no matter how awful, even if they arise from flaws in it.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Since few users ever read sources,
 *    credits must appear in the documentation.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.  Since few users
 *    ever read sources, credits must appear in the documentation.
 *
 * 4. This notice may not be removed or altered.
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <memory.h>
#include "fgetfln.h"
#include "fgetmfs.h"

/* macros */
#define	STREQ(a, b)	(*(a) == *(b) && strcmp((a), (b)) == 0)
#define	STREQN(a, b, n)	((n) == 0 || *(a) == *(b) && strncmp((a), (b), (n)) == 0)
#define STRLEN(s) (sizeof s - 1)		/* s must be a char array */
#define SKIPBLANKS(s) while (*(s) == ' ') (s)++
#define min(a, b) ((a) <= (b)? (a): (b))
#ifndef SLOW
/* modifies s; ++ skips \ in \- */
#define skipesc(s) (((s) = (*(s) == TRESC? ignfontch(s): (s))), \
		    (*(s) == TRESC && *++(s) == '(' && (s)[1] != NUL? \
		     (s)+STRLEN("(b"): (s)))
#endif

/* keywords */
#define STATIC static			/* for dumb debuggers and profilers */

/* constants */
#define YES 1
#define NO 0

#define TRESC '\\'			/* troff escape */
#define NUL '\0'
#define NEWLINE '\n'

/* tunable parameters */
#define MAXSECTLEN 25		/* max length of actual section headings */
#ifndef PAGESIZE
#define PAGESIZE 66
#endif
#define MAXSAVEDLINES (PAGESIZE+20)

/* imports */
extern int optind;
extern char *optarg;
extern char *emalloc();
extern FILE *efopen();

/* exports */
char *progname;
int debug;

/* forwards */
char *ignfontch(), *strsave(), *firstwd(), *parsepara(), *complextag();
char *skipbl(), *parsesection(), *getnxtline(), *parsetext(), *readahead();
#ifdef SLOW
char *skipesc();
#endif

/* private */
struct layout {
	/* indentations */
	int leftmarg;			/* first char of title line */
	int normind;			/* normal body text */
	int indparaind;			/* indented-paragraph */
	int rightmarg;			/* last char of title line */
	/* these are to assist in recognising headers and footers: */
	int lineno, pagesize, headsize, footsize;	/* line numbers */
	char *title, *section;			/* .TH strings */
};
static int savedparas = 0;
static char paramac[] = ".PP\n";

/*
 * main - parse arguments and handle options
 */
main(argc, argv)
int argc;
char *argv[];
{
	int c, errflg = 0;

#ifdef DEBUG
	setbuf(stdout, (char *)NULL);
	mal_debug(3);
#endif
	progname = argv[0];
	while ((c = getopt(argc, argv, "d")) != EOF)
		switch (c) {
		case 'd':
			++debug;
			break;
		default:
			errflg++;
			break;
		}
	if (errflg) {
		(void) fprintf(stderr, "usage: %s [-d] [file]...\n", progname);
		exit(2);
	}

	if (optind >= argc)
		process(stdin, "stdin");
	else
		for (; optind < argc; optind++)
			if (STREQ(argv[optind], "-"))
				process(stdin, "-");
			else {
				FILE *in = efopen(argv[optind], "r");

				process(in, argv[optind]);
				(void) fclose(in);
			}
	exit(0);
}

/*
 * process - process input file
 */
process(in, inname)
FILE *in;
char *inname;
{
	struct layout layout;
	register struct layout *lp = &layout;
	int headsize = 0, incr = 0;
	char *line;

	lp->lineno = 0;
	lp->normind = 0;
	lp->pagesize = PAGESIZE; /* let's pray that page lengths are constant */
	lp->footsize = findfooter(in, lp);	/* reads ahead */
	line = skipbl(in, lp, &headsize);
	if (line != NULL) {
		parsetitle(inname, line, lp);
		(void) printf(".TH %s %s\n", lp->title, lp->section);
		line = skipbl(in, lp, &incr);
		headsize += incr;
		lp->headsize = headsize;
		while (line != NULL)
			line = parsesection(in, inname, line, lp);
	}
}

/*
 * return number of lines from footer to foot.  reads PAGESIZE lines, notes
 * line number of last non-empty line, assumed to be the start of the footer,
 * and pushes back all the non-footer lines it read, in reverse order of
 * reading.  getnxtline reads too far when it sees blank lines, and there
 * are no previously read lines, so use fgets instead.
 */
int
findfooter(in, lp)
register FILE *in;
register struct layout *lp;
{
	register int lineno = 0, linenum;
	register int lastnonempty = PAGESIZE, footsize;
	register char *line;
	char *lines[MAXSAVEDLINES];

	while ((line = fgetms(in)) != NULL && lineno < PAGESIZE) {
		if (line[0] != NEWLINE)
			lastnonempty = lineno;
		lines[lineno++] = line;		/* for later push back */
	}
	footsize = (lineno < PAGESIZE? 0: PAGESIZE - lastnonempty);
	linenum = lineno;
	while (--linenum >= 0) {
		if (linenum < lp->pagesize - footsize)	/* pre-footer? */
			saveline(lines[linenum]);
		free(lines[linenum]);
	}
	return footsize;
}

/*
 * skip blank lines; return non-blank line and store count of lines skipped.
 */
char *
skipbl(in, lp, linenop)
register FILE *in;
register struct layout *lp;
register int *linenop;
{
	register char *line;

	*linenop = 0;
	while ((line = getnxtline(in, lp)) != NULL && line[0] == NEWLINE)
		(*linenop)++;
	return line;
}

/* tear apart the header line and deduce .TH from it (into *lp) */
parsetitle(inname, line, lp)
register char *inname, *line;
register struct layout *lp;
{
	register char *title, *paren, *fontch, *section = NULL;

	lp->leftmarg = countsp(line);
	lp->rightmarg = width(line);
	title = ignfontch(line + lp->leftmarg);
	fontch = strchr(line, TRESC);
	paren = strchr(line, '(');
	if (paren != NULL) {
		*paren = NUL;
		section = paren + 1;
	}
	if (fontch != NULL)
		*fontch = NUL;		/* terminate title */
	if (section == NULL) {
		(void) fprintf(stderr, "%s: in `%s',\n", progname, line);
		botch("no section number in title", inname);
		section = "1";
	} else
		chopat(section, ')');	/* terminate section */
	lp->title = strsave(title);
	lp->section = strsave(section);
}

char *
ignfontch(s)		/* skip a sequence of troff font changes (\fX) */
register char *s;
{
	register int c;
	static char fontch[] = "\\f";
#define FONTCHLEN STRLEN(fontch)

	while (STREQN(s, fontch, FONTCHLEN) && (c = s[FONTCHLEN]) != NUL)
		/* 2-letter fontname? 1+2: ( + fontname; 1: fontname */
		s += FONTCHLEN + (c == '('? 1 + 2: 1);
	return s;
}

int
countsp(s)					/* count leading spaces */
register char *s;
{
	register char *p = s;

	while (*p++ == ' ')
		;
	return p - 1 - s;
}

STATIC
botch(reason, file)
char *reason, *file;
{
	(void) fprintf(stderr, "%s: %s in `%s'\n", progname, reason, file);
}

STATIC char *
strsave(s)
register char *s;
{
	register char *p = emalloc(strlen(s) + 1);

	if (p != NULL)
		(void) strcpy(p, s);
	return p;
}

/*
 * parse a manual section (stop at a section keyword)
 * and emit corresponding troff -man input.
 * line is a buffer, initially containing a (presumed) section keyword.
 * set lp->normind from paragraph indent.
 * skip page footers ("Page n " against leftmarg) and headers.
 */
char *
parsesection(in, inname, line, lp)
FILE *in;
register char *inname, *line;
register struct layout *lp;
{
	register char *keyword, *endkw;
	register int indent, lwid = -1;
	int newpara = YES;
	char *quote;
#ifdef NICE_IDEA_BUT
	static char fontbold[] = "\\fB";
#endif

	indent = countsp(line);
	if (indent != lp->leftmarg) {
		botch("section keyword not against left margin", inname);
		keyword = "BOTCHED";
	} else
		keyword = ignfontch(line + lp->leftmarg);
	chopat(keyword, NEWLINE);
	endkw = strrchr(keyword, TRESC);		/* font change? */
	if (endkw != NULL)
		*endkw = NUL;
	quote = (strchr(keyword, ' ') != NULL? "\"": "");
	(void) printf(".SH %s%s%s\n", quote, keyword, quote);
	savedparas = 0;				/* toss any pending .PP */
	newpara = YES;

	line = getnxtline(in, lp);
	if (lp->normind == 0 && line != NULL)
		lp->normind = countsp(line);
	lp->indparaind = lp->normind;
	while (line != NULL &&
	    ((indent = countsp(line)) > lp->leftmarg ||
#ifdef NICE_IDEA_BUT
	     !STREQN(line, fontbold, STRLEN(fontbold)) ||
#endif
	     line[0] == NEWLINE ||
	     (lwid = width(line)) > MAXSECTLEN)) {	/* HP protection */
		if (line[0] == NEWLINE)		/* paragraph approaching */
			savedparas++;
		else {
			if (lwid < 0)
				lwid = width(line);
			line = parsetext(line, lp, indent, in, newpara, lwid);
		}
		newpara = NO;
		line = getnxtline(in, lp);
		lwid = -1;
	}
	return line;
}

STATIC
chopat(s, c)
char *s;
int c;
{
	register char *at = strchr(s, c);

	if (at != NULL)
		*at = NUL;
}

/*
 * examine a line of text in a section and recognise the start of indented
 * paragraphs.  since tags and indents can be small (1 character each), we
 * have to look ahead to guess reliably(!).
 *
 * N.B.: don't throw away an exdented line! it's probably a table from HP.
 *
 * indented or exdented lines need a .ti before and a .br after (yes, this is
 * a bit of hack but it's a last resort).  short lines not at a paragraph end
 * also need a .br after, where "short line" is defined by "first word on next
 * line would have fit on this line".
 */
char *
parsetext(line, lp, indent, in, newpara, lwid)
register char *line;
register struct layout *lp;
register int indent;
FILE *in;
int newpara, lwid;
{
	register char *text = line + lp->normind;
	int looks_just = NO, exdented = NO;
	char *nxtline;

	nxtline = readahead(in, lp);
	if (savedparas > 0 || newpara)	/* first line of new para? */
		text = parsepara(text, nxtline, indent, newpara, lwid,
			&looks_just, lp);
	else {				/* later line of a paragraph */
		/*
		 * check for current indentation but must eat \ sequences when
		 * counting.
		 */
		if (indent != lp->indparaind)
			(void) fprintf(stderr,
		"%s: line at wrong indentation (%d) for paragraph (%d): %s",
				progname, indent, lp->indparaind, line);
		SKIPBLANKS(text);	/* find para text */
	}
	if (indent < lp->normind) {
		(void) fprintf(stderr, "%s: exdented line: %s", progname, line);
		exdented = YES;
		text = line;		/* don't lose part of the line */
		SKIPBLANKS(text);	/* find start of text */
	}

	/* normal line-by-line treatment */
	if (indent > lp->indparaind || exdented) {
		register int ti = indent - lp->indparaind;

		(void) printf(".ti %c%d\n", (ti < 0? '-': '+'), abs(ti));
	}
	putunjust(line, text, in, lp, looks_just, lwid);
	if (indent > lp->indparaind || exdented ||
	    shortline(nxtline, indent, lwid, lp))
		(void) fputs(".br\n", stdout);
	return line;
}

/*
 * does first word of the next line fit between the end of this line and the
 * right margin, and is the next line indented at least as much as this line?
 */
int
shortline(nxtline, indent, lwid, lp)
register char *nxtline;
int indent, lwid;
struct layout *lp;
{
	return width(firstwd(nxtline)) < lp->rightmarg - lwid &&
	    countsp(nxtline) >= indent;
}

/*
 * start-of-paragraph treatment.  there are traps a-plenty here:  a tag may
 * contain spaces or may be too wide to fit in the alloted gap and thus occupy
 * a line by itself.  again, readahead can resolve these cases.
 */
STATIC char *				/* new text value */
parsepara(text, nxtline, indent, newpara, lwid, ljp, lp)
register char *text, *nxtline;
int indent, newpara, lwid;
int *ljp;
register struct layout *lp;
{
	register char *sp = strchr(text, ' ');
	register int paraind = (sp != NULL? gap(text, sp): 0);
	register int nxtspcs = (nxtline == NULL? 0: countsp(nxtline));
	register char *nsp = NULL;
	int nparaind = 0;

	lp->indparaind = lp->normind;	/* normal indentation with a section */
	savedparas = 0;			/* undo .PP's */
	if (shortline((nxtline == NULL? "": nxtline), indent, lwid, lp) &&
	    nxtspcs > indent) {
		paraind = nxtspcs - lp->normind;
		(void) printf(".TP %d\n", paraind);
		/* no need to emit the tag (text); higher levels will do so */
	} else if (paraind == 0) {	/* no spaces? normal paragraph */
		if (!newpara)		/* not just after .SH */
			putline(paramac);
	} else if (sp == text)		/* starts with space? indented para */
		(void) printf(".IP \"\" %d\n", paraind);
	else if (nxtspcs == lp->indparaind + paraind ||
	    (nsp = complextag(text, nxtspcs, &nparaind, lp)) != NULL) {
		/*
		 * "nxtspcs == lp->indparaind + paraind" means next line has
		 * same indentation as this one, ignoring any tag, so this is a
		 * tagged para.  "nsp != NULL" means the tag contains spaces.
		 */
		if (nsp != NULL) {
			paraind = nparaind;
			sp = nsp;
		}
		puttag(text, sp, paraind);
		text = sp;		/* skip the tag */
		SKIPBLANKS(text);	/* find para text */
		*ljp = looksjust(text, lp, lwid);
	} else {			/* normal paragraph */
		if (!newpara)		/* not just after .SH */
			putline(paramac);
		paraind = 0;		/* ignore presence of spaces */
	}
	lp->indparaind += paraind;
	SKIPBLANKS(text);		/* find para text */
	return text;
}

STATIC char *
complextag(text, nxtspcs, npip, lp)
register char *text;
register int nxtspcs;
register int *npip;
struct layout *lp;
{
	register int nxtind = nxtspcs - lp->normind;
	register char *sp;

	if (nxtind > 0)		/* next line more indented than this one? */
		for (sp = text; (sp = strchr(sp, ' ')) != NULL; sp++) {
			*npip = gap(text, sp);
			if (*npip == nxtind)
				return sp;
		}
	*npip = 0;			/* paranoia */
	return NULL;
}

STATIC
puttag(text, sp, paraind)
char *text, *sp;
int paraind;
{
	register char *tag, *tagsp;

	(void) printf(".TP %d\n", paraind);
	tag = strsave(text);
	tagsp = (sp - text) + tag;
	tagsp[0] = NEWLINE;
	tagsp[1] = NUL;		/* may clobber non-space */
	putline(tag);
	free(tag);
}

/* return pointer to static copy of first word of s */
STATIC char *
firstwd(s)
char *s;
{
	register char *iwd, *wdstart;
	register char savec;
	static char *word;

	for (iwd = s; *iwd == ' '; iwd++)
		;
	for (wdstart = iwd; *iwd != NUL && *iwd != ' ' && *iwd != NEWLINE;
	     iwd++)
		;
	savec = *iwd;
	*iwd = NUL;
	if (word != NULL)
		free(word);
	word = strsave(wdstart);
	*iwd = savec;
	return word;
}

/*
 * compute the width of the gap between the normal left margin (line) within a
 * section and the margin of the indented paragraph (at or after space).
 */
int
gap(line, space)
register char *line, *space;
{
	register int gapsz = 0;

	while (*line != NUL && line < space) {
		line = skipesc(line);
		if (*line != NUL && line < space)
			line++, gapsz++;
	}
	while (*line == ' ')
		line++, gapsz++;
	return gapsz;
}

/*
 * compute the width in printable characters of line.
 * this is a profiling hot spot, so it's been tuned.
 */
STATIC int
width(line)
register char *line;
{
	register int chars = 0, c = *line;

	if (c == NUL || c == NEWLINE)
		return chars;
	for (; ; ) {
		if (c == TRESC) {	/* optimisation */
			line = skipesc(line);
			c = *line;
		}
		if (c == NUL || c == NEWLINE)
			break;		/* NOTE loop exit */
		c = *++line;
		chars++;
	}
	return chars;
}

#ifdef SLOW
/*
 * skip troff font changes completely; for other troff escape sequences, skip
 * them and return a pointer to the last byte.
 */
STATIC char *
skipesc(line)
register char *line;
{
	if (*line == TRESC)		/* optimisation */
		line = ignfontch(line);
	/* ++ skips \ in \- */
	if (*line == TRESC && *++line == '(' && line[1] != NUL)
		line += STRLEN("(b");	/* skip (b in \(bu */
	return line;

}
#endif

/*
 * try to detect if line is justified; if so, unjustify it and emit it.
 * oldline is the entire line; linefrag is where text (after any tag) starts.
 * this ugly hack is somewhat error-prone and is only necessary for
 * awf and other V6-derived man macros that justify lines.
 * to try to minimise the false positives, we read ahead a little
 * to see if the apparent justifiedness persists.
 */
STATIC
putunjust(oldline, linefrag, in, lp, looks_just, lwid)
register char *oldline, *linefrag;
FILE *in;
register struct layout *lp;
int looks_just, lwid;
{
	register char *outfrag = linefrag;

	if (looks_just || looksjust(linefrag, lp, lwid)) {
		register int nxtind, nxtwid;
		char *nxtline;

		nxtline = readahead(in, lp);
		if (nxtline != NULL) {
			nxtind = countsp(nxtline);
			if (nxtind == gap(oldline, linefrag) &&
			    ((nxtwid = width(nxtline)) < lp->rightmarg || /* short */
			     looksjust(nxtline + nxtind, lp, nxtwid))) {
				outfrag = emalloc(strlen(oldline)+1);
				un_just(linefrag, outfrag);
			}
		}
	}
	putline(outfrag);
	if (outfrag != linefrag)
		free(outfrag);
}


/*
 * copy linefrag to unjust, deleting extra spaces (not when preceded by a dot
 * though).
 */
STATIC
un_just(linefrag, unjust)
register char *linefrag, *unjust;
{
	if (linefrag[0] != NUL)
		/* first byte is always okay */
		for (*unjust++ = *linefrag++; *linefrag != NUL;
		     *unjust++ = *linefrag++)
			if (linefrag[-1] != '.' && linefrag[0] == ' ' &&
			    linefrag[1] == ' ')	/* a run of whitespace? */
				while (linefrag[1] == ' ')
					linefrag++;
	*unjust = NUL;
}

STATIC int
looksjust(linefrag, lp, olwid)
register char *linefrag;
register struct layout *lp;
int olwid;
{
	register char *p;

	if (olwid == lp->rightmarg)		/* could be justified */
		for (p = linefrag; (p = strchr(p, ' ')) != NULL; p++)
			if (p[1] == ' ')
				return YES;	/* two spaces in a row */
	return NO;
}


/*
 * write line on stdout, treasuring up .PP's, to be emitted at the first
 * non-.PP line.  treat lines starting with . or ' specially and prepend \&.
 * deals with hyphenation by turning "-" at line end into "-\c".
 */
STATIC
putline(line)
register char *line;
{
	static char hyphen[] = "-\n";
#define HYPHENLEN STRLEN(hyphen)

	if (STREQ(line, paramac))
		savedparas++;
	else {
		register int len, hyphenate = NO;

		for (; savedparas > 0; savedparas--)
			(void) fputs(paramac, stdout);
		if (line[0] == '.' || line[0] == '\'')
			(void) fputs("\\&", stdout);	/* invisible char */
		len = strlen(line);
		if (len >= HYPHENLEN && STREQ(&line[len-HYPHENLEN], "-\n")) {
			hyphenate = YES;
			line[len-1] = NUL;		/* stomp nl */
		}
		(void) fputs(line, stdout);
		if (hyphenate)
			(void) fputs("\\c\n", stdout);	/* kludge */
	}
}

/*
 * this rather elaborate machinery provides pushback of lines.
 */

static int blanks = 0;
static struct pushback {
	char *svline;
	int svlen;		/* bytes actually saved */
} pb[MAXSAVEDLINES];
static int nextfreeline = 0;

STATIC
saveline(line)
char *line;
{
	register struct pushback *pbp;

	if (nextfreeline >= MAXSAVEDLINES)
		botch("can't push back any more lines", "");
	pbp = &pb[nextfreeline++];
	/* don't use strsave here to avoid redundant strlen's */
	pbp->svlen = strlen(line) + 1;
	pbp->svline = emalloc(pbp->svlen);
	if (pbp->svline == NULL)
		botch("can't allocate memory to save a line", "");
	(void) strcpy(pbp->svline, line);
}

STATIC char *					/* malloced */
returnsaved()
{
	register char *line;

	if (blanks > 0) {			/* any saved blank lines? */
		--blanks;
		line = strsave("\n");
	} else if (nextfreeline > 0) {
		register struct pushback *pbp = &pb[--nextfreeline];

		line = pbp->svline;
		/* free(pbp->svline); */
		pbp->svline = NULL;
	} else
		botch("returnsaved called inappropriately - can't happen", "");
	return line;
}

char *
readahead(in, lp)		/* read ahead one line */
FILE *in;
struct layout *lp;
{
	char *buf = getnxtline(in, lp);

	if (buf != NULL)
		saveline(buf);		/* put back */
	return buf;
}

/*
 * read the next non-footer, non-header line into "line" from "in".
 * skip page footers ("Page n " against leftmarg) and headers.
 * this is probably subtly wrong for the first footer.
 */
char *
getnxtline(in, lp)
FILE *in;
register struct layout *lp;
{
	static char *line;			/* line needing to be freed */

#ifdef FREELINES /* a worthy sentiment, but it interferes with readahead */
	if (line != NULL) {
		free(line);
		line = NULL;
	}
#endif
	if (blanks > 0 || nextfreeline > 0) {	/* any saved (blank) lines? */
		line = returnsaved();
		return line;
	}
	while ((line = fgetms(in)) != NULL) {
		lp->lineno++;
		if (lp->lineno >= lp->pagesize)
			lp->lineno = 0;		/* wrap around */

		if (lp->lineno >= lp->pagesize - lp->footsize ||
		    lp->lineno < lp->headsize)
			;			/* skip footers and headers */
		else if (line[0] == NEWLINE)
			++blanks;
		else {				/* normal text */
			if (blanks > 0) {
				saveline(line); /* for later; return blank */
				free(line);
				line = returnsaved();
			}
			return line;		/* NOTE loop exit */
		}
	}
	/* at eof */
	if (blanks > 0) {
		--blanks;
		line = strsave("\n");
		return line;
	} else
		return NULL;
}
